
/*
 *  block 0
 *    bootstrap
 *  block 1
 *    superblock
 *  block 2
 *    inode 1
 *    inode 2     rootdir
 *    inode 3
 *    inode 4
 *    inode 5
 *    inode 6
 *    inode 7
 *    inode 8
 *  block 3
 *    inode 9
 *    ...
 *  block s_isize
 *    datablock s_isize
 *  block s_fsize - 1
 *    last datablock
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/fblk.h"
#include "../h/conf.h"
#include "../h/buf.h"
#include "../h/inode.h"
#include "../h/ino.h"
#include "../h/dir.h"
#include "../h/proc.h"

char size_fs_blocks[(
	BSIZE >= sizeof(struct filsys) &&
	BSIZE >= sizeof(struct fblk)   &&
	BSIZE == sizeof(struct dinode) * INOPB
) + 1] = "x";

/*
 * alloc will obtain the next available
 * free disk block from the free list of
 * the specified device.
 * The super block has up to NICFREE remembered
 * free blocks; the last of these is read to
 * obtain NICFREE more . . .
 *
 * no space on dev x/y -- when
 * the free list is exhausted.
 */
struct buf *
alloc(mp)
	struct mount *mp;
{
	daddr_t bno;
	register struct buf *bp;
	register struct filsys *fp = &mp->m_filsys;

	while (fp->s_flock)
		sleep(&fp->s_flock, PINOD);

	do {
		if (fp->s_nfree <= 0)
			goto nospace;
		if (fp->s_nfree > NICFREE) {
			fserr(mp, "bad free count");
			goto nospace;
		}
		bno = fp->s_free[--fp->s_nfree];
		if (bno == 0)
			goto nospace;
	} while (badblock(mp, bno));

	if (fp->s_nfree <= 0) {
		fp->s_flock++;
		bp = bread(mp->m_dev, bno);
		if ((bp->b_flags&B_ERROR) == 0) {
			fp->s_nfree = bp->b_un.b_fblk->df_nfree;
			bcopy(bp->b_un.b_fblk->df_free, fp->s_free, sizeof(fp->s_free));
		}
		brelse(bp);
		fp->s_flock = 0;
		wakeup(&fp->s_flock);
		if (fp->s_nfree <=0)
			goto nospace;
	}
	bp = getblk(mp->m_dev, bno);
	clrbuf(bp);
	fp->s_fmod = 1;
	return (bp);

nospace:
	fp->s_nfree = 0;
	fserr(mp, "disk full");
	errno = ENOSPC;
	return (NULL);
}

/*
 * place the specified disk block
 * back on the free list of the
 * specified device.
 */
free(mp, bno)
	struct mount *mp;
	daddr_t bno;
{
	register struct filsys *fp = &mp->m_filsys;
	register struct buf *bp;

	fp->s_fmod = 1;
	while (fp->s_flock)
		sleep(&fp->s_flock, PINOD);

	if (badblock(mp, bno))
		return;

	if (fp->s_nfree <= 0) {
		fp->s_nfree = 1;
		fp->s_free[0] = 0;
	}
	if (fp->s_nfree >= NICFREE) {
		fp->s_flock++;
		bp = getblk(mp->m_dev, bno);
		bp->b_un.b_fblk->df_nfree = fp->s_nfree;
		bcopy(fp->s_free, bp->b_un.b_fblk->df_free, sizeof(fp->s_free));
		fp->s_nfree = 0;
		bwrite(bp);
		fp->s_flock = 0;
		wakeup(&fp->s_flock);
	}
	fp->s_free[fp->s_nfree++] = bno;
	fp->s_fmod = 1;
}

/*
 * Check that a block number is in the
 * range between the I list and the size
 * of the device.
 * This is used mainly to check that a
 * garbage file system has not been mounted.
 *
 * bad block on dev x/y -- not in range
 */
badblock(mp, bn)
	struct mount *mp;
	daddr_t bn;
{
	register struct filsys *fp = &mp->m_filsys;

	if (bn < fp->s_isize
	 || bn >= fp->s_fsize) {
		fserr(mp, "bad block");
		return (1);
	}
	return (0);
}

/*
 * Allocate an unused I node
 * on the specified device.
 * Used with file creation.
 * The algorithm keeps up to
 * NICINOD spare I nodes in the
 * super block. When this runs out,
 * a linear search through the
 * I list is instituted to pick
 * up NICINOD more.
 */
struct inode *
ialloc(mp)
	struct mount *mp;
{
	register struct filsys *fp = &mp->m_filsys;
	register struct inode *ip;
	register struct buf *bp;
	int i;
	struct dinode *dp;
	ino_t ino, inobas;
	int first;
	daddr_t adr;

loop:
	while (fp->s_ilock)
		sleep(&fp->s_ilock, PINOD);

	if (fp->s_ninode > 0) {
		ino = fp->s_inode[--fp->s_ninode];
		ip = iget(mp, ino);
		if (ip == NULL)
			return (NULL);
		if (ip->i_mode == 0) {
			fp->s_fmod = 1;
			bzero(ip->i_addr, sizeof(ip->i_addr));
			return (ip);
		}
		/*
		 * Inode was allocated after all.
		 * Look some more.
		 */
		iput(ip);
		goto loop;
	}
	fp->s_ilock++;

	if (fp->s_nbehind < 4 * NICINOD) {
		first = 1;
		ino = fp->s_lasti;
		if (itoo(ino))
			panic("ialloc");
		adr = itod(ino);
	} else {
fromtop:
		fp->s_nbehind = 0;
		first = 0;
		ino = 1;
		adr = SUPERB+1;
	}
	for( ; adr < fp->s_isize; ++adr) {
		inobas = ino;
		bp = bread(mp->m_dev, adr);
		if (bp->b_flags & B_ERROR) {
			brelse(bp);
			ino += INOPB;
			continue;
		}
		dp = bp->b_un.b_dino;
		for (i = 0; i < INOPB; ++i) {
			if (dp->di_mode)
				goto cont;
			for (ip = inode; ip < &inode[NINODE]; ++ip)
				if (ip->i_number == ino && ip->i_mp == mp)
					goto cont;
			fp->s_inode[fp->s_ninode++] = ino;
			if (fp->s_ninode >= NICINOD)
				break;
		cont:
			ino++;
			dp++;
		}
		brelse(bp);
		if (fp->s_ninode >= NICINOD)
			break;
	}
	if (fp->s_ninode < NICINOD && first)
		goto fromtop;

	fp->s_lasti = inobas;
	fp->s_ilock = 0;
	wakeup(&fp->s_ilock);
	if (fp->s_ninode > 0)
		goto loop;

	fserr(mp, "out of inodes");
	errno = ENOSPC;
	return (NULL);
}

/*
 * Free the specified I node
 * on the specified device.
 * The algorithm stores up
 * to NICINOD I nodes in the super
 * block and throws away any more.
 */
ifree(mp, ino)
	struct mount *mp;
	ino_t ino;
{
	register struct filsys *fp = &mp->m_filsys;

	if (fp->s_ilock)
		return;

	if (fp->s_ninode >= NICINOD) {
		if (fp->s_lasti > ino)
			fp->s_nbehind++;
		return;
	}
	fp->s_inode[fp->s_ninode++] = ino;
	fp->s_fmod = 1;
}

/*
 * update is the internal name of
 * 'sync'. It goes through the disk
 * queues to initiate sandbagged IO;
 * goes through the I nodes to write
 * modified nodes; and it goes through
 * the mount table to initiate modified
 * super blocks.
 */
update()
{
	register struct inode *ip;
	struct mount *mp;
	struct buf *bp;
	struct filsys *fp;

	if (updlock)
		return;
	updlock = 1;

	for (mp = mount; mp < &mount[NMOUNT]; ++mp)
		if (mp->m_dev != NODEV) {
			fp = &mp->m_filsys;
			if (fp->s_fmod == 0
			 || fp->s_ilock
			 || fp->s_flock
			 || fp->s_ronly)
				continue;

			bp = getblk(mp->m_dev, SUPERB);
			if (bp->b_flags & B_ERROR)
				continue;
			fp->s_fmod = 0;
			bcopy(fp, bp->b_un.b_addr, sizeof(*fp));
			bwrite(bp);
		}
	for (ip = inode; ip < &inode[NINODE]; ++ip)
		if (ip->i_count && (ip->i_flag&ILOCK)==0) {
			ip->i_flag |= ILOCK;
			ip->i_count++;
			iupdat(ip);
			iput(ip);
		}
	updlock = 0;
	bflush(NODEV);
}

