
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/inode.h"

/*
 * Convert a user supplied
 * file descriptor into a pointer
 * to a file structure.
 * Only task is to check range
 * of the descriptor.
 */
struct file *
getf(f)
	unsigned f;
{
	struct file *fp;

	if (f < NOFILE) {
		fp = u.u_ofile[f];
		if (fp)
			return (fp);
	}
	errno = EBADF;
	return (NULL);
}

/*
 * Internal form of close.
 * Decrement reference count on
 * file structure.
 * Also make sure the pipe protocol
 * does not constipate.
 *
 * Decrement reference count on the inode following
 * removal to the referencing file structure.
 * Call device handler on last close.
 */
closef(fp)
	register struct file *fp;
{
	struct inode *ip;
	int flag, mode;
	dev_t dev;
	register int (*cfunc)();

	if (fp == NULL)
		return;

	if (fp->f_count != 1) {
		fp->f_count--;
		return;
	}
	flag = fp->f_flag;
	ip   = fp->f_inode;
	dev  = (dev_t) ip->i_rdev;
	mode = ip->i_mode;

	plock(ip);
	fp->f_count = 0;
	fp->f_inode = NULL;

	if (flag & FPIPE) {
		ip->i_mode &= ~(IREAD|IWRITE);
		wakeup((caddr_t)ip+1);
		wakeup((caddr_t)ip+2);
	}
	iput(ip);

	switch (mode&IFMT) {
	default:
		return;

	case IFCHR:
		cfunc = cdevsw[major(dev)].d_close;
		break;

	case IFBLK:
		cfunc = bdevsw[major(dev)].d_close;
		break;
	}

	if (cfunc != noop) {
		for (fp = file; fp < &file[NFILE]; ++fp)
			if (fp->f_inode == ip)
				return;

		(*cfunc)(dev, flag);
	}
}

/*
 * openf called to allow handler of special files
 * to initialize and validate before actual IO.
 */
openf(fp)
	struct file *fp;
{
	struct inode *ip = fp->f_inode;
	dev_t dev = (dev_t) ip->i_rdev;
	unsigned maj = major(dev);
	int istty;

	switch (ip->i_mode&IFMT) {

	case IFCHR:
		/* ttymaj is 128 or something like that, not part of cdevsw[] */

		if (maj == ttymaj) {
			if (curproc->p_ctty == NULL)
				goto bad;
			iput(ip);
			ip = curproc->p_ctty;
			ip->i_count++;
			fp->f_inode = ip;
			break;
		}
		if (maj >= nchrdev)
			goto bad;

		/* devices which provide 'controlling tty' can speak now */

		istty = 0;
		(*cdevsw[maj].d_open)(dev, fp->f_flag, &istty);
		if (errno)
			return;

		if (curproc->p_ctty == NULL && (istty & 1)) {
			ip->i_count++;
			curproc->p_ctty = ip;
		}
		break;

	case IFBLK:
		if (maj >= nblkdev)
			goto bad;
		(*bdevsw[maj].d_open)(dev, fp->f_flag);
	}
	return;

bad:
	errno = ENXIO;
}

/*
 * Check mode permission on inode pointer.
 * Mode is mix of READ, WRITE and EXEC.
 * In the case of WRITE, the read-only status of the file system is checked.
 * Also in WRITE, prototype text segments cannot be written.
 */
access(ip, mode)
	register struct inode *ip;
	register mode;
{
	if (mode & IWRITE) {
		if (ip->i_mp->m_filsys.s_ronly) {
			errno = EROFS;
			return (0);
		}
		if (ip->i_flag & ITEXT) {
			errno = ETXTBSY;
			return (0);
		}
	}
	if ((ip->i_mode & mode) != mode) {
		errno = EACCES;
		return (0);
	}
	return (1);
}

/*
 * Allocate a user file descriptor.
 */
ufalloc()
{
	register i;

	for (i = 0; i < NOFILE; ++i)
		if (u.u_ofile[i] == NULL) {
			u.u_pofile[i] = 0;
			return (i);
		}
	errno = EMFILE;
	return (-1);
}

/*
 * Allocate a user file descriptor
 * and a file structure.
 * Initialize the descriptor
 * to point at the file structure.
 *
 * no file -- if there are no available
 * 	file structures.
 */
struct file *
falloc(fdp)
	int *fdp;
{
	register struct file *fp;
	register i;

	i = ufalloc();
	if (i < 0)
		return (NULL);

	for (fp = file; fp < &file[NFILE]; ++fp)
		if (fp->f_count == 0) {
			fp->f_count = 1;
			fp->f_un.f_offset = 0;

			*fdp = i;
			u.u_ofile[i] = fp;
			u.u_pofile[i] = 0;
			return (fp);
		}
	errno = ENFILE;
	return (NULL);
}

