
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/conf.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/buf.h"
#include "../h/ioctl.h"

sys_read(fd, buf, cc)
{
	return (rdwr(FREAD, fd, buf, cc));
}

sys_write(fd, buf, cc)
{
	return (rdwr(FWRITE, fd, buf, cc));
}

/*
 * common code for read and write calls:
 * check permissions, set base, count, and offset,
 * and switch out to readi, writei, or pipe code.
 * also handle virtual reads and writes.
 */
int
rdwr(mode, fd, buf, cc)
	char *buf;
{
	register struct file *fp;
	register struct inode *ip;

	if (cc < 0) {
		errno = EINVAL;
		return;
	}
	fp = getf(fd);
	if (fp == NULL)
		return;

	if ((fp->f_flag&mode) == 0) {
		errno = EBADF;
		return;
	}
	u.u_count = cc;
	u.u_base = buf;
	u.u_segflg = 1;

	if (fp->f_flag&FPIPE) {
		if (mode == FREAD)
			readp(fp);
		else
			writep(fp);
	} else {
		ip = fp->f_inode;

		if ((ip->i_mode&(IFCHR&IFBLK)) == 0)
			plock(ip);

		if (mode == FWRITE && (fp->f_flag&FAPPEND))
			fp->f_un.f_offset = ip->i_size;

		u.u_offset = fp->f_un.f_offset;

		if (mode == FREAD)
			readi(ip, fp->f_flag);
		else
			writei(ip, fp->f_flag);

		if ((ip->i_mode&(IFCHR&IFBLK)) == 0)
			prele(ip);

		fp->f_un.f_offset += cc - u.u_count;
	}
	return (cc - u.u_count);
}

sys_open(fname, mode, perm)
	char *fname;
{
	struct inode *ip;
	struct file *fp;
	int fd;

	if ((fp = falloc(&fd)) == NULL)
		return (-1);

	if (mode & FCREAT) {
		ip = namei(fname, 1);
		if (ip == NULL) {
			if (errno)
				goto bad;
			ip = maknode((perm & IPERM) | IFREG);
			mode &= ~FTRUNC;
		} else {
			if (mode & FEXCL) {
				errno = EEXIST;
				goto bad;
			}
			mode &= ~FCREAT;
		}
	} else {
		ip = namei(fname, 0);
	}
	if (ip == NULL)
		goto bad;

	if ((mode & FCREAT) == 0) {
		if (!access(ip, mode & (FREAD|FWRITE)))
			goto bad;
		if ((mode&FWRITE) && (ip->i_mode&IFMT) == IFDIR) {
			errno = EISDIR;
			goto bad;
		}
	}
	if (mode & FTRUNC)
		itrunc(ip);
	prele(ip);

	fp->f_flag = mode & FMASK;
	fp->f_inode = ip;

	/* open character or block device.
	 * the inode might change (from pseudo-files like /dev/tty)  */

	openf(fp);

	ip = fp->f_inode;

	if (errno == 0)
		return (fd);

bad:
	u.u_ofile[fd] = NULL;
	fp->f_count = 0;
	fp->f_inode = NULL;
	if (ip)
		iput(ip);

	return (-1);
}

sys_close(fd)
{
	register struct file *fp;

	fp = getf(fd);
	if (fp == NULL)
		return (-1);

	u.u_ofile[fd] = NULL;
	closef(fp);
}

off_t
sys_lseek(fd, off, sbase)
	off_t off;
{
	register struct file *fp;

	fp = getf(fd);
	if (fp == NULL)
		return (-1L);

	if (fp->f_flag&FPIPE) {
		errno = ESPIPE;
		return (-1L);
	}
	switch (sbase) {
	case 1:
		off += fp->f_un.f_offset;
		break;
	case 2:
		off += fp->f_inode->i_size;
		break;
	}
	fp->f_un.f_offset = off;

	return (off);
}

sys_mknod(fname, fmode, dev)
	char *fname;
{
	struct inode *ip;

	ip = namei(fname, 1);
	if (ip) {
		errno = EEXIST;
		goto out;
	}
	if (errno)
		return;

	ip = maknode(fmode);
	if (ip == NULL)
		return;

	if (fmode & (IFCHR&IFBLK))
		ip->i_rdev = (dev_t)dev;

out:
	iput(ip);
}

sys_link(target, linkname)
	char *target, *linkname;
{
	register struct inode *ip, *xp;

	ip = namei(target, 0);
	if (ip == NULL)
		return;

	if (ip->i_nlink == LINK_MAX) {
		errno = EMLINK;
		goto out;
	}
	/*
	 * Unlock to avoid possibly hanging the namei.
	 * Sadly, this means races. (Suppose someone
	 * deletes the file in the meantime?)
	 * Nor can it be locked again later
	 * because then there will be deadly
	 * embraces.
	 */
	prele(ip);
	xp = namei(linkname, 1);
	if (xp) {
		errno = EEXIST;
		iput(xp);
		goto out;
	}
	if (errno)
		goto out;

	if (u.u_pdir->i_mp != ip->i_mp) {
		iput(u.u_pdir);
		errno = EXDEV;
		goto out;
	}
	wdir(ip->i_number);

	if (errno == 0) {
		ip->i_nlink++;
		ip->i_flag |= ICHG;
	}
out:
	iput(ip);
}

/*
 * Hard to avoid races here, especially
 * in unlinking directories.
 */
sys_unlink(fname)
	char *fname;
{
	register struct inode *ip, *pp;

	pp = namei(fname, 2);
	if (pp == NULL)
		return;

	/*
	 * Check for unlink(".")
	 * to avoid hanging on the iget
	 */
	if (pp->i_number == u.u_dent.d_ino) {
		ip = pp;
		ip->i_count++;
	} else {
		ip = iget(pp->i_mp, u.u_dent.d_ino);
		if (ip == NULL)
			goto out1;
	}
	/*
	 * Don't unlink a mounted file.
	 */
	if (ip->i_mp != pp->i_mp) {
		errno = EBUSY;
		goto out;
	}
	u.u_dent.d_ino = 0;

	u.u_offset -= sizeof(struct direct);
	u.u_count = sizeof(struct direct);
	u.u_base = &u.u_dent;
	u.u_segflg = 0;
	writei(pp, 0);

	if (ip->i_nlink) {
		ip->i_nlink--;
		ip->i_flag |= ICHG;
	}
out:
	iput(ip);
out1:
	iput(pp);
}

/* XXX staging... error checking...
 * XXX */

sys_mkdir(path, perm)
	char *path;
{
	static char dots[DIRSIZ+1] = ".."; /* dots+1 is "." */

	struct inode *ip;
	struct direct dir[2];

	/* namei(, 1) and maknode() work together with u.u_pdir */

	ip = namei(path, 1);
	if (ip) {
		errno = EEXIST;
		goto out;
	}
	if (errno)
		return;

	ip = u.u_pdir;

	/* dotdot, parent of 'path' */

	if (ip->i_nlink == LINK_MAX) {
		errno = EMLINK;
		goto out;
	}
	ip->i_nlink++;
	dir[1].d_ino = ip->i_number;

	ip = maknode((perm & IPERM) | IFDIR);
	if (ip == NULL)
		return;

	/* dot, 'path' itself */

	ip->i_nlink++;
	dir[0].d_ino = ip->i_number;

	/* write contents */

	bcopy(dots + 1, dir[0].d_name, DIRSIZ);
	bcopy(dots + 0, dir[1].d_name, DIRSIZ);

	u.u_offset = 0;
	u.u_count = sizeof(dir);
	u.u_base = dir;
	u.u_segflg = 0;
	writei(ip, 0);
out:
	iput(ip);
}

sys_chdir(path)
	char *path;
{
	register struct inode *ip;

	ip = namei(path, 0);
	if (ip == NULL)
		return;

	if ((ip->i_mode&IFMT) != IFDIR) {
		errno = ENOTDIR;
		goto bad;
	}
	if (!access(ip, IEXEC))
		goto bad;

	prele(ip);
	plock(curproc->p_cdir);
	iput(curproc->p_cdir);

	curproc->p_cdir = ip;
	return;

bad:
	iput(ip);
}

sys_chmod(fname, fmode)
	char *fname;
{
	register struct inode *ip;

	if ((ip = namei(fname, 0)) == NULL)
		return;

	ip->i_mode &= ~IPERM;
	ip->i_mode |= fmode&IPERM;
	ip->i_flag |= ICHG;
	iput(ip);
}

sys_fchmod(fd, fmode)
{
	struct file *fp;
	register struct inode *ip;

	if ((fp = getf(fd)) == NULL)
		return;

	ip = fp->f_inode;
	plock(ip);
	ip->i_mode &= ~IPERM;
	ip->i_mode |= fmode&IPERM;
	ip->i_flag |= ICHG;
	prele(ip);
}

sys_ioctl(fd, cmd, arg)
{
	struct file *fp;
	struct inode *ip;
	dev_t dev;

	if ((fp = getf(fd)) == NULL)
		return;

	switch (cmd) {
	case FIOCLEX:
		u.u_pofile[fd] |= EXCLOSE;
		return;

	case FIONCLEX:
		u.u_pofile[fd] &= ~EXCLOSE;
		return;
	}

	ip = fp->f_inode;
	dev = ip->i_rdev;

	if ((ip->i_mode & IFMT) != IFCHR) {
		errno = ENOTTY;
		return;
	}
	return (*cdevsw[major(dev)].d_ioctl)(dev, cmd, arg, fp->f_flag);
}

